; ******************************************************************************
;
; FAT32 MBR Boot Strap Code for
; The Micro (Disk) Operating System for 6502 Microprocessors
;
; M/OS-65 System I
; by Joerg Walke
;
; Developed for the Junior Computer ][
;
; First implementation 6.7.2023
; by Joerg Walke
;
; Assembled with A65
;
; M/OS-65 MBR Boot Strap Code for FAT32 Devices
; Final Stable Version
;
; This code is freely available under the Creative Commons Attribution 4.0
; International license.
; See https://creativecommons.org/licenses/by/4.0/
;
; ******************************************************************************

PSTR            EQU     $EA

NUM32      	EQU   	$F8     	; low 32 bit number byte
SUM32           EQU     $FC             ; low 32 bit number byte
NUM16      	EQU   	$FC     	; low 16 bit number byte
NUM8            EQU     $FE             ; 8 bit number byte

YSAV            EQU     $FD             ; Temp Saved Device ID

STOL      	EQU   	$E8     	; Store address Low
STOH      	EQU   	$E9     	; Store address High

BLKBUFL		EQU	$DC
BLKBUFH		EQU	$DD

RESET           EQU     $E000
SWITCH_TO_RAM   EQU     $E006
COUT            EQU     $E052
LFOUT           EQU     $E05A
CIN             EQU     $E047
STROUT          EQU     $E083
HEXOUT          EQU     $E091

CMDDEV          EQU     $E0BA

;SD_RD_LBLK      EQU     $F462

MOUNTED_DEVS    EQU     $0400           ; Table of mounted devices
BOOT_PART       EQU     MOUNTED_DEVS    ; Boot Medium Descriptor

; Constants *******************

Sectors_Per_Cluser EQU  $060D
Reserved_Sectors   EQU  $060E
Number_of_FATs     EQU  $0610
Num_Root_Dir_Entr  EQU  $0611
Num_Log_Sectors    EQU  $0613
Medium_Descr_Byte  EQU  $0615
Sectors_Per_FAT    EQU  $0624
Dir_Start_Cluster  EQU  $062C

DIR_BLK_BUF     EQU     $0200

BOOT_CODE       EQU     $2000

D_PART_START    EQU     BOOT_PART
D_PART_SIZE     EQU     BOOT_PART+$04
D_DEV_ID        EQU     BOOT_PART+$08
D_MEDIUM_DESCR  EQU     BOOT_PART+$09
D_FAT_TYPE      EQU     BOOT_PART+$0A
D_SECT_PER_CLST EQU     BOOT_PART+$0B
D_NUM_OF_FAT    EQU     BOOT_PART+$0C
D_NUM_ROOT_DIR  EQU     BOOT_PART+$0D
D_START_CLUSTER EQU     BOOT_PART+$10
D_START_FAT1    EQU     BOOT_PART+$14
D_START_FAT2    EQU     BOOT_PART+$18
D_START_DIR     EQU     BOOT_PART+$1C

FAT12_Type      EQU     $02
FAT16_Type      EQU     $01
FAT32_Type      EQU     $00

CMD_READ	EQU	34

;NO_BOOT_DEV     EQU     $E496

;****************************************************************************************
;****************************************************************************************

PROG_START	EQU	$065A		; Program Start Address
USE_XMODEM	EQU	0		; 1 use XModem, 0 don't use XModem

;****************************************************************************************

	IF	USE_XMODEM EQ 1

		ORG	PROG_START-2

		DB    	LOW  PROG_START
		DB	HIGH PROG_START

	ENDI

;****************************************************************************************

		ORG	PROG_START	; the program start address

;****************************************************************************************
;****************************************************************************************

; 600: EB 58 90 4A 43 4F 53 20 20 20 20
; 400: 00 20 00 00
;3000: A2 00 A0 04 20 92 F4 A9 00 2A 20 91 E0 60  - WRITE BLOCK
;3010: A2 00 A0 04 20 5F F4 A9 00 2A 20 91 E0 60  - READ BLOCK


MAIN            LDA     YSAV                ; get current device id
                STA     D_DEV_ID            ; save device id to descriptor
                ;LDA     Sectors_Per_Cluser  ; get Sectors_Per_Cluser
                ;STA     D_SECT_PER_CLST     ; save Sectors_Per_Cluser to descriptor
                LDA     Medium_Descr_Byte   ; get Medium_Descriptor_Byte
                STA     D_MEDIUM_DESCR      ; save Medium_Descriptor_Byte to descriptor
SET_FAT32       LDX     #FAT32_Type         ; set FAT32 type ID
                STX     D_NUM_ROOT_DIR      ; set number of root dir entries to 0
                STX     D_NUM_ROOT_DIR+1
SET_FAT_TYPE    STX     D_FAT_TYPE          ; store FAT type
CALC_FAT1_POS   LDX     #LOW  D_PART_START  ; load pointer to Partition_Start_Sector
                LDY     #HIGH D_PART_START
                JSR     LOAD_32             ; load Start_Sector value
                LDX     Reserved_Sectors    ; load low byte of Reserved_Sectors
                LDY     Reserved_Sectors+1  ; load high byte of Reserved_Sectors
                JSR     ADD_32_16           ; add Reserved_Sectors to Start_Sector
                LDX     #LOW  D_START_FAT1
                LDY     #HIGH D_START_FAT1
                JSR     STORE_32            ; store result to Start_Of_FAT1 Pointer
                LDX     #LOW  Sectors_Per_FAT ; load low byte of Sectors_Per_FAT
                LDY     #HIGH Sectors_Per_FAT ; load high byte of Sectors_Per_FAT
                JSR     LOAD_S32
                JSR     ADD_32_32           ; add Sectors_Per_FAT to Start_Of_FAT1 Pointer
                LDX     #LOW  D_START_FAT2
                LDY     #HIGH D_START_FAT2
                JSR     STORE_32            ; and store result to Start_Of_FAT2 Pointer
                LDY     Number_of_FATs      ; get Number_of_FATs
                STY     D_NUM_OF_FAT        ; save Number_of_FATs to descriptor
                DEY
                BNE     LOOP_CALC_ROOT
                JMP     BOOT_ERR_MSG        ; error if number of FATs is zero
LOOP_CALC_ROOT  JSR     ADD_32_32           ; add Sectors_Per_FAT to Start_Of_FAT1 Pointer
                DEY                         ; until Number_Of_FATs is 0
                BNE     LOOP_CALC_ROOT
SET_CLSTR_START LDX     #LOW  D_START_CLUSTER
                LDY     #HIGH D_START_CLUSTER
                JSR     STORE_32            ; and store result as Start_Of_Clusters to descriptor
                LDX     #$03
SET_DIR_START   LDA     Dir_Start_Cluster,X ; copy directory start cluster into volume descriptor
                STA     D_START_DIR,X
                STA     NUM32,X             ; and NUM32[0:3]
                DEX
                BPL     SET_DIR_START
                
; calc LBA of first directory block ********************************************

                SEC
                LDA     #$02
                TAX
                SBC     NUM32
                STA     NUM32
                LDA     #$00
SUB_OFFS        SBC     NUM32,X
                STA     NUM32,X
                DEX
                BPL     SUB_OFFS

                LDA     Sectors_Per_Cluser  ; get Sectors_Per_Cluser
                STA     D_SECT_PER_CLST     ; load number of sectors per clusters
                
                LSR     A
                BEQ     ADD_CLUSTER         ; more than one sector?
LOOP1           LDX     #$00                ; yes, multiply directory start cluster
                LDY     #$04                ; with number of sectors per cluster
                CLC
LOOP2           ROL     NUM32,X
                INX
                DEY
                BNE     LOOP2
                LSR     A
                BNE     LOOP1
ADD_CLUSTER     LDX     #LOW  D_START_CLUSTER
                LDY     #HIGH D_START_CLUSTER
                JSR     LOAD_S32            ; load start cluster LBA
                JSR     ADD_32_32           ; and add it to directory start LBA

; load first root directory block **********************************************

LOAD_ROOT_DIR   LDA     #LOW  DIR_BLK_BUF   ; set read buffer to DIR_BLK_BUFF
                STA     BLKBUFL
                STA     PSTR
                LDA     #HIGH DIR_BLK_BUF
                STA     BLKBUFH
                STA     PSTR+1
                LDX     #LOW  NUM32
                LDY     #HIGH NUM32
                JSR     DEV_RD_LBLK          ; read first block of root directory

READ_DIR_ENTRY  LDY     #$00
READ_FIRST_CHR  LDA     (PSTR),Y            ; read first char of filename
                BEQ     BOOT_ERR_MSG        ; is entry empty? Yes, BOOT.SYS not found
READ_NAME       LDA     (PSTR),Y
                CMP     #$20
                BCC     GET_NEXT_ENTRY
                CMP     #$61
                BCC     COMPARE_CHR         ; is it a upper case char?
                AND   	#$DF	 	    ; no, convert to uppercase char
COMPARE_CHR     CMP     BOOT_FILE,Y
                BNE     GET_NEXT_ENTRY
                INY
                CPY     #$0B
                BCC     READ_NAME

                LDA     #LOW  BOOT_CODE     ;
                STA     BLKBUFL
                LDA     #HIGH BOOT_CODE     ;
                STA     BLKBUFH
                
                SEC
                LDY     #$1A
                LDA     (PSTR),Y
                SBC     #$02
                STA     STOL
                INY
                LDA     (PSTR),Y
                SBC     #$00
                STA     STOH
                LDA     D_SECT_PER_CLST
CALC_OFFS       LSR     A
                BEQ     CALC_CLUSTER
                ASL     STOL
                ROL     STOH
                JMP     CALC_OFFS
CALC_CLUSTER    LDX     STOL
                LDY     STOH
                JSR     ADD_32_16
                LDX     #LOW  NUM32
                LDY     #HIGH NUM32
                JSR     DEV_RD_LBLK         ; read first block of BOOT.SYS
                JSR     SWITCH_TO_RAM       ; switch $B000..$DFFF to RAM bank
                JMP     BOOT_CODE           ; jump to boot code
                

DEV_RD_LBLK     LDA     #CMD_READ
                JMP     CMDDEV

GET_NEXT_ENTRY  CLC
                LDA     #32
                ADC     PSTR
                STA     PSTR
                LDA     #$00
                ADC     PSTR+1
                STA     PSTR+1
                CMP     #$04
                BCC     READ_DIR_ENTRY
                
BOOT_ERR_MSG    LDA     #LOW MSG            ; no BOOT.SYS file found
                STA     PSTR                ; load error message
                LDA     #HIGH MSG
                STA     PSTR+1
                JSR     STROUT              ; and print it
                JSR     CIN                 ; wait for any key pressed
                JMP     RESET               ; and reset
                
; **** Load a 32 Bit Value Into SUM32 ******************************************
;
; Input:  X,Y = Ptr[LO:HI] to 32 Bit Integer
;
; ******************************************************************************

LOAD_S32        STX     STOL
                LDX     #$07
                BNE     LOAD_32_1

; **** Load a 32 Bit Value Into NUM32 ******************************************
;
; Input:  X,Y = Ptr[LO:HI] to 32 Bit Integer
;
; ******************************************************************************

LOAD_32         STX     STOL
                LDX     #$03
LOAD_32_1       STY     STOH
                LDY     #$03
LOOP_LOAD_32    LDA     (STOL),Y
                STA     NUM32,X
                DEX
                DEY
                BPL     LOOP_LOAD_32
                RTS

; Input:  X,Y = Ptr[LO:HI] to 32 Bit Integer

STORE_32        STX     STOL
                STY     STOH
                LDY     #$03
LOOP_STORE_32   LDA     NUM32,Y
                STA     (STOL),Y
                DEY
                BPL     LOOP_STORE_32
                RTS
                
; Input X,Y = 16 Bit Integer, NUM32[0:3] = 32 Bit Integer - NUM32 = UInt[X,Y]+NUM32
                
ADD_32_16       STX     SUM32
                STY     SUM32+1
                LDX     #$00
                STX     SUM32+2
                STX     SUM32+3
                
; Input SUM32[0:3] = 32 Bit Integer, NUM32[0:3] = 32 Bit Integer  - NUM32 = SUM32+NUM32
                
ADD_32_32       CLC
                LDX     #$00
                PHP
ADD_LOOP        PLP
                LDA     SUM32,X
                ADC     NUM32,X
                STA     NUM32,X
                PHP
                INX
                CPX     #$04
                BNE     ADD_LOOP
                PLP
                RTS
                
; Boot Messages
                
MSG             DB      $0D,$0D
                TEXT    " No valid boot disk"
                DB      $0D
                TEXT    " Press any key to restart"
                DB      $00
BOOT_FILE       TEXT    "BOOT    SYS"
                ORG     $07FE
                
TAG             DB      $55,$AA
                END